use strict;
use warnings;

package HTTP::Server::Multiplex;

use HTTP::Server::VirtualHost;
use HTTP::Server::VirtualHost::LocalHost;
use HTTP::Server::Connection;

use IO::Multiplex    ();
use IO::Socket::INET ();
use Sys::Hostname    qw(hostname);
use POSIX            qw(setsid);
use English          qw(-no_match_vars);
use POSIX            qw(setuid setgid sigprocmask
                        SIGINT SIG_BLOCK SIG_UNBLOCK);
use Fcntl;
use File::Spec       ();
use Socket           qw(inet_aton AF_INET);

use Log::Report 'httpd-multiplex', syntax => 'SHORT';

###

=chapter NAME
HTTP::Server::Multiplex - single process multi serve HTTP daemon

=chapter SYNOPSIS

  # see examples directory in the distribution!
  use HTTP::Server::Multiplex;
  my $daemon = HTTP::Server::Multiplex->new
    ( daemon     => {}
    , connection => {}
    , vhosts     => {}
    );

  $daemon->addVirtualHost('My::VHost');

  $daemon->run;

=chapter DESCRIPTION
This full stand-alone HTTP daemon serves all requests from B<one single
process>.  See the L</DETAILS> about the back-ground, advantages and
backdraws of this approach.  You should read that section at least once
before you start.

B<This is the first release: please test and report bugs and wish-lists!>

=chapter METHODS

=c_method new OPTIONS

=requires deamon     HASH-of-OPTIONS
For OPTIONS, see M<_configDaemon()>

=requires connection SOCKET|HASH-of-OPTIONS|ARRAY
For OPTIONS, see M<_configNetwork()>.  You can provide an ARRAY
of SOCKETs or socket configuration OPTIONS, which will be handled
in parallel.

=option  vhosts      VHOST|HASH-of-OPTIONS|PACKAGE|ARRAY
=default vhosts      <HTTP::Server::VirtualHost::Default>
For OPTIONS, see M<addVirtualHost()>.  Provide one or an ARRAY of
virtual host configurations, either by M<HTTP::Server::VirtualHost>
objects or by the OPTIONS to create such objects.
=cut

my $singleton;
sub new(@)
{   my $class = shift;
    my $args  = @_==1 ? shift @_ : {@_};

    error __x"you can only create one {pkg} object per program"
        if $singleton++;   # only one IO::Multiplexer

    (bless {}, $class)->init($args);
}

sub _to_list($) { ref $_[0] eq 'ARRAY' ? @{$_[0]} : $_[0] }
sub init($)
{   my ($self, $args) = @_;

    my $mux = $self->{HSM_mux} = IO::Multiplex->new;
    $mux->set_callback_object($self);

    foreach my $conn (_to_list delete $args->{connection})
    {   trace "setting up connection";
        $self->_configNetwork($mux, $conn);
    }

    $self->{HSM_vhosts} = {};
    foreach my $vhost (_to_list delete $args->{vhosts})
    {   trace "setting up virtual host";
        $self->addVirtualHost($vhost);
    }

    trace "setting up daemon";
    $self->_configDaemon(delete $args->{daemon});

    error __x"Unknown option for ::Multiplex::new(): {names}"
      , names => [keys %$args]
         if keys %$args;
    $self;
}

=method _configNetwork SOCKET|HASH-of-OPTIONS
Set-up a listener waiting for new connections to arrive.  This is an
internal helper routine for M<new()>.

=option  listen INTEGER
=default listen SO_MAXCONN

=option  host   STRING
=default host   <all interfaces>
Needed to get the interface IP addresses from.

=option  port   INTEGER
=default port   80
=cut

sub _configNetwork($$)
{   my ($self, $mux, $config) = @_;
    my $socket;

    if(UNIVERSAL::isa($config, 'IO::Socket'))
    {   $socket = $config;
    }
    elsif(not UNIVERSAL::isa($config, 'HASH'))
    {   error __x"connection configuration not a socket not HASH";
    }
    else
    {   my $host = $config->{host} || '0.0.0.0';
        my $port = $config->{port} || 80;
        $socket = IO::Socket::INET->new
          ( LocalAddr => $host
          , Listen    => 5
          , LocalPort => $port
          , Reuse     => 1  # to be able to restart without loss of service
                            # not yet implemented
          );

        defined $socket
           or fault __x"unable to create socket for {host} port {port}"
                , host => $host, port => $port;

        trace 'created server socket '.$socket->sockhost.':'.$port;
    }

    $mux->listen($socket);
}

=method _configDaemon OPTIONS

=option  detach    BOOLEAN
=default detach    <false>
When true, the daemon will detach from the terminal and run in
the background.  STDIN, STDOUT, STDERR will get closed; errors
go to syslog via M<Log::Report::dispatcher()>.

=option  server_id STRING
=default server_id <shows hostname and some versions>

=option  user      STRING
=default user      C<$ENV{USER}> or C<$EUID>
The username under which the daemon will be run.

If you are the super-user and you want this daemon to run as such
(which is dangerous!), then you must explicitly provide the name of the
super-user username.

=option  group     STRING
=default group     C<$EGID>
The group(s) under which the daemon will be run.

=option  pid_file  FILENAME
=default pid_file  <undef>
=cut

sub _configDaemon($)
{   my ($self, $config) = @_;
    my @daemon_headers;

    my $id;
    if(exists $config->{server_id})
    {   $id = $config->{server_id};
    }
    else
    {   no strict; no warnings;
        $id = hostname . " ".__PACKAGE__." $VERSION, "
            . "IO::Multiplex $IO::Multiplex::VERSION";
    }
    push @daemon_headers, Server => $id if defined $id;
    HTTP::Server::Connection->setDefaultHeaders(@daemon_headers);

    $EUID!=0 || defined $config->{user}
        or error __"running daemon as root is dangerous: specify other user";

    my $user   = $config->{user} || $ENV{USER} || $EUID;
    my $uid    = $user =~ m/\D/ ? getpwnam($user) : $user;
    defined $uid
        or error __x"user {name} does not exist", name => $user;
    $self->{HSM_uid} = $uid;

    my @groups = split ' ', ($config->{group} || $EGID);
    my @gid;
    foreach my $group (@groups)
    {   my $gid = $group =~ m/\D/ ? getgrnam($group) : $group;
        defined $gid
            or error __x"group {name} does not exist", name => $group;
        push @gid, $gid;
    }
    $self->{HSM_gid} = join ' ', @gid;

    $self->{HSM_pidfn} = $config->{pid_file};
    $self;
}

sub _daemonize()
{   my $self = shift;

    my ($uid, $gid) = @$self{'HSM_uid', 'HSM_gid'};
    if($uid ne $EUID)
    {   setuid $uid
            or fault __x"cannot switch to user-id {uid}", uid => $uid;
        trace "switch to user $uid";
    }
    if($gid ne $EGID)
    {   setgid $gid
            or fault __x"cannot switch to group-id {gid}", gid => $gid;
        trace "switch to group $gid";
    }

    $self->{HSM_detach}
        or return $self;

    my $pidfile = $self->{HSM_pidfn};
    if(defined $pidfile)
    {   sysopen PID, $pidfile, O_EXCL|O_CREAT|O_WRONLY|O_TRUNC
            or fault __x"cannot write to pid_file {fn}", fn => $pidfile;
    }

    trace "close standard error dispatcher";
    dispatcher close => 'PERL';    # no die/warn output

    trace "closing standard file-handles";
    open STDIN,  '<', File::Spec->devnull;
    open STDOUT, '>', File::Spec->devnull;
    open STDERR, '>', File::Spec->devnull;

    trace "process into the background";
    my $sigset = POSIX::SigSet->new(SIGINT);
    sigprocmask SIG_BLOCK, $sigset
        or fault "cannot block SIGINT for fork";

    my $pid    = fork;
    defined $pid
        or fault "cannot fork into background";

    sigprocmask SIG_UNBLOCK, $sigset
        or fault "cannot unblock SIGINT after fork";

    if($pid > 0)
    {   # Parent process
        if($pidfile)
        {   print PID "$pid\n";
            close PID or fault "cannot write pid-file {fn}", fn => $pidfile;
        }

        return $self;
    }

    # Child process
    close PID if $pidfile;

    setsid;

    $self;
}

#-------------
=section Accessors

=method mux
Returns the core multiplexer object, an M<IO::Multiplex>.
=cut

sub mux() {shift->{HSM_mux}}

#-------------
=section Daemon control

=method run
Start the daemon.
=cut

sub run()
{   my $self = shift;

    unless(keys %{$self->{HSM_vhosts}})
    {   trace "creating default vhost 'localhost' because no explicit vhost";
        $self->addVirtualHost(HTTP::Server::VirtualHost::LocalHost->new);
    }

    $self->_daemonize;

    info __x"http daemon start, user {uid} group {gid}"
      , uid => $EUID, gid => $EGID;

    $self->mux->loop;
}

#-------------
=section Virtual host administration

=method addVirtualHost VHOST|PACKAGE|HASH-of-OPTIONS|OPTIONS
Adds a new virtual host to the knowledge of the daemon.  Can be used
at run-time.  See M<HTTP::Server::VirtualHost::new()> for OPTIONS.
The added virtual host object is returned.

=cut

sub addVirtualHost(@)
{   my $self   = shift;
    my $config = @_==1 ? shift : {@_};
    my $vhost;
    if(UNIVERSAL::isa($config, 'HTTP::Server::VirtualHost'))
    {   $vhost = $config;
    }
    elsif(!ref $config && $config =~ m/\:\:/)
    {   eval "require $config";
        die $@ if $@;
        $vhost = $config->new;
    }
    elsif(not UNIVERSAL::isa($config, 'HASH'))
    {   error __x"virtual configuration not a valid object not HASH";
    }
    else
    {   $vhost = HTTP::Server::VirtualHost->new($config);
    }

    $self->{HSM_vhosts}{$_} = $vhost
        for $vhost->name, $vhost->aliases;
    $vhost;
}

=method removeVirtualHost VHOST|NAME|ALIAS
Remove all name and alias registrations for the indicated virtual host.
Silently ignores non-existing vhosts.  The removed virtual host object
is returned.
=cut

sub removeVirtualHost($)
{   my ($self, $id) = @_;
    my $vhost = UNIVERSAL::isa($id, 'HTTP::Server::VirtualHost') ? $id
              : $self->virtualHost($id);
    defined $vhost or return;

    delete $self->{HSM_vhosts}{$_}
        for $vhost->name, $vhost->aliases;
    $vhost;
}

=method virtualHost NAME
=cut

sub virtualHost($) { $_[0]->{HSM_vhosts}{$_[1]} }

#-------------------
#section Multiplexer

sub mux_connection($$)
{   my ($self, $mux, $fh) = @_;
    my $client = HTTP::Server::Connection->new($mux, $fh, $self);
    $mux->set_callback_object($client, $fh);
}

sub dnslookup($$$)
{   my ($self, $conn, $ip, $where) = @_;
    my $host = $self->{HSM_cache}{$ip} ||=
        # must be changed into async lookup!
        gethostbyaddr inet_aton($ip), AF_INET;
    $$where  = $host;
    info $conn->id." $ip is $host";
}

#-------------------

=chapter DETAILS

=section HTTP::Daemon/mod_perl differ from this approach

The LWP network library is very solid and provides a full HTTP/1.1
daemon named M<HTTP::Daemon>.  The logic of that daemon was used to
create the code for this module.  Both LWP and Apache's mod_perl start
many processes, for each requesting client one.  What we do here is to
have only one process to serve many clients at the same time.

HTTP::Daemon and mod_perl are based on processes and threads to handle
requests in parallel.  The advantage is that disturbances and delays in
handling one client's request are not brothering the other processes, the
other clients.  As disadvantage, it quite hard to share user session
information and do caching.  Solutions are found in databases and the
mem-cache daemon to share data, and use locking to synchronize.

This M<HTTP::Server::Multiplex> module uses only one process to handle
all requests, by serving many client connections together.  The base
is laid in M<IO::Multiplex>, which is a smart C<select(2)> call (you may
need to read-up on that Operating System feature).  The single process
spends all its time handling one request, until IO has to be done.
When waiting for that IO to happen, it will handle available request
from other connections.

Advantages of this approach: no heavy forking, no complex synchronizing
between processes and very simple caching.  Most importantly: the whole
code is a critical section which avoid the need for locking.  Very fast
and very simple -in concept.

Disadvantages are also plentifold: you have to be careful with any IO
to use the IO::Multiplex select loop, and busy-waits (for reading files,
acquiring locks, database output, hostname lookups, sleep) are blocking
all other clients.  Any bug in your software can break the daemon,
and therewith the whole service.

=section Features

The following common needs for http servers are implemented:

=over 4

=item HTTP/1.1 (and 1.0 and older)
Multiple requests can use one connection. Requests can arrive asynchronously
from the processing, queuing-up while processed in order.

=item Virtual Hosts
Configure multiple websites, handled by this single daemon.  See
M<HTTP::Server::VirtualHost>.  Each website (virtual host) has its own
directories where the information is taken from, each with a set access
restrictions (allow/deny), rewrite rules, location, etc.  See
M<HTTP::Server::Directory>

=item Critical section
With forking servers, it is difficult to synchronize between threads and
processes: you have to lock or use databases to create critical sections.
In this implementation, the whole program runs in one big critical
section, unless you do IO. Of course, each processing should take little
time, to avoid long response delays for other connections.

=item Asynchronous processing
In case you task does difficult I/O or long computation, you can start
a seperate process with M<HTTP::Server::Connection::async()>.  Be aware
that this comes with a penalty.

=item Sessions
The same user can have multiple connections to the daemon, which use a
single session definition.  This way, the cache of the user's information
cannot get out-of-sync.  See M<HTTP::Server::Session>.

=back

=subsection Missing features

=over 4
=item CGI
Asynchronous execution of external scripts, according to the CGI
protocol.  This is not too hard to implement, it just takes some
time.
=back

=section Microsoft Windows Limitations

Microsofts POSIX implementation does support the select() call, which
is the base of M<IO::Multiplex> used as event loop in the daemon.
However, that implementation does not support the use of FILE and PIPE
handles in the select() as seen on UNIX systems: it only supports
sockets.  As a result, the readFile, writeFile and async methods of
M<HTTP::Server::Connection> will (probably) not work.  Please, users of
Windows: explain me how to implement work-arounds.

=section General Misconceptions

People apparently think that executing requests in parallel processes
is faster than executing all requests in one process; this cannot be
true (in a single processor system): processing takes computational effort
and there is only one processor to run these tasks.  Parallel processes
do not run in parallel, but interleaved.  It is the same amount of work
to be done, so runs in comparible total time.

The most important difference between the one-process and the parallel
implementation, is that the latter can sleep much easier: waiting for
disk, interrupts and such in one parallel thread does not hold-up the
other threads.  In the one-process implementation, we need to code just
like graphical interfaces work: in small fragments, triggered by (file
IO) events.

When you implement your "CGI" logic within the virtual host frame-work
of this daemon, your code gets compiled before the daemon starts
accepting requests.  The same, of course, for mod_perl.  Once running,
it is quite hard to fill a time-slice of availability of the processor;
the OS interleaves parallel processes with rather long time-spans of
(usually) around on tenth of a second.  You can do an enormous amount
of work in such a long time-slice.  Most tasks, however, will not consume
the whole span, because they have to wait for more data from disk,
some event, or are completed.  Concluding that an IO-driven approach is
rarely giving slower "interactive" response to clients than a parallel
implementation.  At least in theory, when implemented correctly.

=cut

1;
